@@ -63,7 +63,7 @@ try:
 except ImportError:
     pass
 
-from collections import defaultdict
+from collections import defaultdict, namedtuple
 from datetime import date
 from email.message import Message
 from gettext import gettext as _
@@ -115,6 +115,8 @@ SIGNAL_STOP_REQUEST = False
 # messages to be logged only once
 logged_msgs = set()  # type: AbstractSet[str]
 
+NEVER_PIN = -32768
+
 
 class LoggingDateTime:
     """The date/time representation for the dpkg log file timestamps"""
@@ -141,12 +143,31 @@ class NoAllowedOriginError(ValueError):
     pass
 
 
+PkgPin = namedtuple('PkgPin', ['pkg', 'priority'])
+PkgFilePin = namedtuple('PkgFilePin', ['id', 'priority'])
+
+
 class UnattendedUpgradesCache(apt.Cache):
 
-    def __init__(self, rootdir, allowed_origins):
-        # type: (str, List[str]) -> None
+    def __init__(self, rootdir):
+        # type: (str) -> None
         self._cached_candidate_pkgnames = set()  # type: Set[str]
-        self.allowed_origins = allowed_origins
+
+        self.allowed_origins = get_allowed_origins()
+        logging.info(_("Allowed origins are: %s"),
+                     ", ".join(self.allowed_origins))
+
+        self.blacklist = apt_pkg.config.value_list(
+            "Unattended-Upgrade::Package-Blacklist")
+        logging.info(_("Initial blacklist: %s"), " ".join(self.blacklist))
+
+        self.whitelist = apt_pkg.config.value_list(
+            "Unattended-Upgrade::Package-Whitelist")
+        self.strict_whitelist = apt_pkg.config.find_b(
+            "Unattended-Upgrade::Package-Whitelist-Strict", False)
+        logging.info(_("Initial whitelist (%s): %s"),
+                     "strict" if self.strict_whitelist else "not strict",
+                     " ".join(self.whitelist))
         apt.Cache.__init__(self, rootdir=rootdir)
 
         # pre-heat lazy-loaded modules to avoid crash on python upgrade
@@ -164,6 +185,150 @@ class UnattendedUpgradesCache(apt.Cache):
             logging.debug("Using %s regexp to find running kernel packages",
                           self.running_kernel_pkgs_regexp.pattern)
 
+    def find_better_version(self, pkg):
+        # type (apt.Package) -> apt.package.Version
+        if pkg.is_installed and pkg.versions[0] > pkg.installed:
+            logging.debug(
+                "Package %s has a higher version available, checking if it is "
+                "from an allowed origin and is not pinned down.", pkg.name)
+            for v in pkg.versions:
+                if pkg.installed < v \
+                   and pkg.installed.policy_priority <= v.policy_priority \
+                   and is_in_allowed_origin(v, self.allowed_origins):
+                    return v
+        return None
+
+    def find_kept_packages(self, dry_run):
+        # type: (bool) -> KeptPkgs
+        """ Find kept packages not collected already """
+
+        kept_packages = KeptPkgs(set)
+        if dry_run:
+            logging.info(_("The list of kept packages can't be calculated in "
+                           "dry-run mode."))
+            return kept_packages
+        for pkg in self:
+            better_version = self.find_better_version(pkg)
+            if better_version:
+                logging.info(self.kept_package_excuse(pkg._pkg,
+                                                      self.blacklist,
+                                                      self.whitelist,
+                                                      self.strict_whitelist,
+                                                      better_version))
+                kept_packages.add(pkg, better_version, self)
+        return kept_packages
+
+    def kept_package_excuse(self, pkg,          # apt.Package
+                            blacklist,          # type: List[str]
+                            whitelist,          # type: List[str]
+                            strict_whitelist,   # type: bool
+                            better_version      # type: apt.package.Version
+                            ):
+        # type: (...) -> str
+        """ Log the excuse the package is kept back for """
+        if pkg.selected_state == apt_pkg.SELSTATE_HOLD:
+            return _("Package %s is marked to be held back.") % pkg.name
+        elif is_pkgname_in_blacklist(pkg.name, blacklist):
+            return _("Package %s is blacklisted.") % pkg.name
+        elif whitelist:
+            if strict_whitelist:
+                if not is_pkgname_in_whitelist(pkg.name, whitelist):
+                    return (_(
+                        "Package %s is not on the strict whitelist.")
+                            % pkg.name)
+                else:
+                    if not is_pkgname_in_whitelist(pkg.name, whitelist):
+                        return (_(
+                            "Package %s is not whitelisted and it is not a"
+                            " dependency of a whitelisted package.")
+                                % pkg.name)
+        elif not any([o.trusted for o in better_version.origins]):
+            return _("Package %s's origin is not trusted.") % pkg.name
+        return (_("Package %s is kept back because a related package"
+                  " is kept back or due to local apt_preferences(5).")
+                % pkg.name)
+
+    def pinning_from_regex_list(self, regexps, priority):
+        # type: (List[str], int) -> List[PkgPin]
+        """ Represent blacklist as Python regexps as list of pkg pinnings"""
+
+        pins = []  # type: List[PkgPin]
+        for regex in regexps:
+            if python_regex_is_posix(regex):
+                pins.append(PkgPin('/' + regex + '/', priority))
+            else:
+                # Python regex is not also an equivalent POSIX regexp.
+                # This is expected to be rare. Go through all the package names
+                # and pin all the matching ones.
+                for pkg in self._cache.packages:
+                    if re.match(regex, pkg.name):
+                        pins.append(PkgPin(pkg.name, priority))
+        return pins
+
+    def pinning_from_config(self):
+        # type: () -> List[Union[PkgPin, PkgFilePin]]
+        """ Represent configuration as list of pinnings
+
+            Assumes self.allowed_origins to be already set.
+        """
+
+        pins = []  # type: List[Union[PkgPin, PkgFilePin]]
+
+        # mark not allowed origins with 'never' pin
+        for pkg_file in self._cache.file_list:  # type: ignore
+            if not is_allowed_origin(pkg_file, self.allowed_origins):
+                # Set the magic 'never' pin on not allowed origins
+                logging.debug("Marking not allowed %s with %s pin", pkg_file,
+                              NEVER_PIN)
+                pins.append(PkgFilePin(pkg_file.id, NEVER_PIN))
+            # TODO(rbalint) pin not trusted origins with NEVER_PIN
+            elif self.strict_whitelist:
+                # set even allowed origins to -1 and set individual package
+                # priorities up later
+                pins.append(PkgFilePin(pkg_file.id, -1))
+
+        # mark blacklisted packages with 'never' pin
+        pins.extend(self.pinning_from_regex_list(  # type: ignore
+            self.blacklist, NEVER_PIN))
+        # set priority of whitelisted packages to high
+        pins.extend(self.pinning_from_regex_list(  # type: ignore
+            self.whitelist, 900))
+        if self.strict_whitelist:
+            policy = self._depcache.policy
+            # pin down already pinned packages which are not on the whitelist
+            # to not install locally pinned up packages accidentally
+            for pkg in self._cache.packages:
+                if pkg.has_versions \
+                   and policy.get_candidate_ver(pkg) > -1:  # type: ignore
+                    # the pin is higher than set for allowed origins, thus
+                    # there is extra pinning configuration
+                    if not is_pkgname_in_whitelist(pkg.name, self.whitelist):
+                        pins.append(PkgPin(pkg.name, NEVER_PIN))
+
+        return pins
+
+    def apply_pinning(self, pins):
+        # type: (List[Union[PkgPin, PkgFilePin]]) -> None
+        """ Apply the list of pins """
+
+        policy = self._depcache.policy
+        pkg_files = {f.id: f for f in self._cache.file_list}  # type: ignore
+        for pin in pins:
+            logging.debug("Applying pinning: %s" % str(pin))
+            if isinstance(pin, PkgPin):
+                policy.create_pin('Version', pin.pkg, '*',  # type: ignore
+                                  pin.priority)
+            elif isinstance(pin, PkgFilePin):
+                logging.debug("Applying pin %s to package_file: %s"
+                              % (pin.priority, str(pkg_files[pin.id])))
+                policy.set_priority(pkg_files[pin.id],  # type: ignore
+                                    pin.priority)
+
+    def open(self, progress=None):
+        apt.Cache.open(self, progress)
+        # apply pinning generated from unattended-upgrades configuration
+        self.apply_pinning(self.pinning_from_config())
+
     def adjust_candidate(self, pkg):
         # type: (apt.Package) -> bool
         """ Adjust origin and return True if adjustment took place
@@ -216,10 +381,6 @@ class UnattendedUpgradesCache(apt.Cache):
             logging.debug("pkg %s is on hold" % pkg.name)
             return
 
-        # adjust candidates in advance if needed
-        for pkg_name in self._cached_candidate_pkgnames:
-            self.adjust_candidate(self[pkg_name])
-
         if function == apt.package.Package.mark_upgrade \
                 and not pkg.is_upgradable:
             if not apt_pkg.config.find_b("Unattended-Upgrade::Allow-downgrade",
@@ -229,7 +390,8 @@ class UnattendedUpgradesCache(apt.Cache):
                 function = apt.package.Package.mark_install
         marking_succeeded = self.call_checked(function, pkg, **kwargs)
 
-        if not marking_succeeded:
+        if not marking_succeeded or \
+           not check_changes_for_sanity(self, desired_pkg=pkg):
             logging.debug("falling back to adjusting %s's dependencies"
                           % pkg.name)
             self.clear()
@@ -416,21 +578,15 @@ class Unlocked:
         # type: () -> None
         try:
             apt_pkg.pkgsystem_unlock_inner()
-        except AttributeError:
-            try:
-                apt_pkg.pkgsystem_unlock()
-            except Exception:
-                # earlier python-apt used to leak lock
-                logging.warning("apt_pkg.pkgsystem_unlock() failed due to not "
-                                "holding the lock but trying to continue")
-                pass
+        except Exception:
+            # earlier python-apt used to leak lock
+            logging.warning("apt_pkg.pkgsystem_unlock() failed due to not "
+                            "holding the lock but trying to continue")
+            pass
 
     def __exit__(self, exc_type, exc_value, exc_tb):
         # type: (object, object, object) -> None
-        try:
-            apt_pkg.pkgsystem_lock_inner()
-        except AttributeError:
-            apt_pkg.pkgsystem_lock()
+        apt_pkg.pkgsystem_lock_inner()
 
 
 class KeptPkgs(defaultdict):
@@ -438,32 +594,16 @@ class KeptPkgs(defaultdict):
     Packages to keep by highest allowed pretty-printed origin
 
     """
-    def add(self, pkg, cache):
-        # type: (apt.Package, UnattendedUpgradesCache) -> None
-        allowed_origins = cache.allowed_origins
-        for origin in ver_in_allowed_origin(pkg, allowed_origins).origins:
-            if is_allowed_origin(origin, allowed_origins):
+    def add(self, pkg,  # type: apt.Package
+            version,    # type: apt.package.Version
+            cache       # type: UnattendedUpgradesCache
+            ):
+        # type: (...) -> None
+        for origin in version.origins:
+            if is_allowed_origin(origin, cache.allowed_origins):
                 self[origin.origin + " " + origin.archive].add(pkg.name)
                 return
 
-    def pop_upgradable(self, cache):
-        # type: (UnattendedUpgradesCache) -> List[apt.Package]
-        upgradable = []
-        empty_sets = set()
-        for name, pkg_set in self.items():
-            remove_from_set = set()
-            for pkg_name in pkg_set:
-                pkg = cache[pkg_name]
-                if pkg.marked_install or pkg.marked_upgrade:
-                    remove_from_set.add(pkg_name)
-                    upgradable.append(pkg)
-            pkg_set -= remove_from_set
-            if not pkg_set:
-                empty_sets.add(name)
-        for empty_set in empty_sets:
-            del(self[empty_set])
-        return upgradable
-
 
 class UnattendedUpgradesResult:
     """
@@ -474,7 +614,7 @@ class UnattendedUpgradesResult:
                  success,                 # type: bool
                  result_str="",           # type: str
                  pkgs=[],                 # type: List[str]
-                 pkgs_kept_back=None,     # type: KeptPkgs
+                 pkgs_kept_back=KeptPkgs(set),     # type: KeptPkgs
                  pkgs_removed=[],         # type: List[str]
                  pkgs_kept_installed=[],  # type: List[str]
                  update_stamp=False       # type: bool
@@ -484,8 +624,6 @@ class UnattendedUpgradesResult:
         self.result_str = result_str
         self.pkgs = pkgs
         self.pkgs_kept_back = pkgs_kept_back
-        if pkgs_kept_back is None:
-            self.pkgs_kept_back = KeptPkgs(set)
         self.pkgs_removed = pkgs_removed
         self.pkgs_kept_installed = pkgs_kept_installed
         self.update_stamp = update_stamp
@@ -514,7 +652,7 @@ def signal_handler(signal, frame):
 
 
 def log_once(msg):
-    #type: (str) -> None
+    # type: (str) -> None
     global logged_msgs
     if msg not in logged_msgs:
         logging.info(msg)
@@ -654,7 +792,7 @@ def get_allowed_origins():
 
 
 def match_whitelist_string(whitelist, origin):
-    # type: (str, apt.package.Origin) -> bool
+    # type: (str, Union[apt.package.Origin, apt_pkg.PackageFile]) -> bool
     """
     take a whitelist string in the form "origin=Debian,label=Debian-Security"
     and match against the given python-apt origin. A empty whitelist string
@@ -700,6 +838,12 @@ def match_whitelist_string(whitelist, origin):
     return res
 
 
+def python_regex_is_posix(expression):
+    # type: (str) -> bool
+    """ Returns if the Python regex is also an equivalent POSIX regex """
+    return re.match("^[a-zA-Z0-9\\^\\$\\+\\.]*$", expression) is None
+
+
 def cache_commit(cache,           # type: apt.Cache
                  logfile_dpkg,    # type: str
                  verbose,         # type: bool
@@ -714,11 +858,7 @@ def cache_commit(cache,           # type: apt.Cache
         iprogress = LogInstallProgress(logfile_dpkg, verbose)
 
     try:
-        if hasattr(apt_pkg, "pkgsystem_lock_inner"):
-            res = cache.commit(install_progress=iprogress)
-        else:
-            with Unlocked():
-                res = cache.commit(install_progress=iprogress)
+        res = cache.commit(install_progress=iprogress)
         cache.open()
     except SystemError as e:
         error = e
@@ -743,8 +883,6 @@ def upgrade_normal(cache, logfile_dpkg, verbose):
 
 def upgrade_in_minimal_steps(cache,            # type: UnattendedUpgradesCache
                              pkgs_to_upgrade,  # type: List[str]
-                             blacklist,        # type: List[str]
-                             whitelist,        # type: List[str]
                              logfile_dpkg="",  # type: str
                              verbose=False,    # type: bool
                              ):
@@ -753,9 +891,6 @@ def upgrade_in_minimal_steps(cache,            # type: UnattendedUpgradesCache
 
     res = True
 
-    # double check any changes we do
-    allowed_origins = get_allowed_origins()
-
     # to upgrade contains the package names
     to_upgrade = set(pkgs_to_upgrade)
     for pkgname in upgrade_order(to_upgrade, cache):
@@ -786,8 +921,7 @@ def upgrade_in_minimal_steps(cache,            # type: UnattendedUpgradesCache
 
         # double check that we are not running into side effects like
         # what could have been caused LP: #1020680
-        if not check_changes_for_sanity(cache, allowed_origins, blacklist,
-                                        whitelist):
+        if not check_changes_for_sanity(cache):
             logging.info("While building minimal partition: "
                          "cache has not allowed changes")
             cache.clear()
@@ -827,7 +961,12 @@ def upgrade_in_minimal_steps(cache,            # type: UnattendedUpgradesCache
 
 
 def is_allowed_origin(origin, allowed_origins):
-    # type: (apt.package.Origin, List[str]) -> bool
+    # type: (Union[apt.package.Origin, apt_pkg.PackageFile], List[str]) -> bool
+
+    # local origin is allowed by default
+    if origin.component == 'now' and origin.archive == 'now' and \
+       not origin.label and not origin.site:
+        return True
     for allowed in allowed_origins:
         if match_whitelist_string(allowed, origin):
             return True
@@ -847,14 +986,6 @@ def is_in_allowed_origin(ver, allowed_origins):
 def ver_in_allowed_origin(pkg, allowed_origins):
     # type: (apt.Package, List[str]) -> apt.package.Version
     for ver in pkg.versions:
-        # ignore versions that the user marked with priority < 100
-        # (and ensure we have a python-apt that supports this)
-        try:
-            if ver.policy_priority < 100:
-                logging.debug("ignoring ver %s with priority < 0" % ver)
-                continue
-        except AttributeError:
-            pass
         if is_in_allowed_origin(ver, allowed_origins):
             # leave as soon as we have the highest new candidate
             return ver
@@ -883,8 +1014,8 @@ def is_pkgname_in_whitelist(pkgname, whitelist):
     return False
 
 
-def is_pkg_change_allowed(pkg, blacklist, whitelist):
-    # type: (apt.Package, List[str], List[str]) -> bool
+def is_pkg_change_allowed(pkg, blacklist, whitelist, strict_whitelist):
+    # type: (apt.Package, List[str], List[str], bool) -> bool
     if is_pkgname_in_blacklist(pkg.name, blacklist):
         logging.debug("pkg %s package has been blacklisted" % pkg.name)
         return False
@@ -892,8 +1023,6 @@ def is_pkg_change_allowed(pkg, blacklist, whitelist):
     # whitelist, most people will want the relaxed whitelist
     # that whitelists a package but pulls in the package
     # dependencies
-    strict_whitelist = apt_pkg.config.find_b(
-        "Unattended-Upgrade::Package-Whitelist-Strict", False)
     if strict_whitelist and \
        not is_pkgname_in_whitelist(pkg.name, whitelist):
 
@@ -950,11 +1079,9 @@ def upgrade_order(to_upgrade, cache):
     return sorted(upgrade_set_sizes, key=upgrade_set_sizes.get)
 
 
-def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist,
-                             desired_pkg=None):
-    # type: (apt.Cache, List[str], List[str], List[str], apt.Package) -> bool
-    sanity_check_result = sanity_problem(cache, allowed_origins, blacklist,
-                                         whitelist, desired_pkg=None)
+def check_changes_for_sanity(cache, desired_pkg=None):
+    # type: (UnattendedUpgradesCache, apt.Package) -> bool
+    sanity_check_result = sanity_problem(cache, desired_pkg)
     if sanity_check_result is None:
         return True
     else:
@@ -964,9 +1091,8 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist,
         return False
 
 
-def sanity_problem(cache, allowed_origins, blacklist, whitelist,
-                   desired_pkg=None):
-    # type: (apt.Cache, List[str], List[str], List[str], apt.Package) -> str
+def sanity_problem(cache, desired_pkg):
+    # type: (UnattendedUpgradesCache, apt.Package) -> str
     if cache._depcache.broken_count != 0:
         return ("there are broken packages in the cache")
     # If there are no packages to be installed they were kept back
@@ -981,9 +1107,12 @@ def sanity_problem(cache, allowed_origins, blacklist, whitelist,
             # origin so its good enough if we have a single trusted one
             if not any([o.trusted for o in pkg.candidate.origins]):
                 return ("pkg %s is not from a trusted origin" % pkg.name)
-            if not is_in_allowed_origin(pkg.candidate, allowed_origins):
+            if not is_in_allowed_origin(pkg.candidate, cache.allowed_origins):
                 return ("pkg %s is not in an allowed origin" % pkg.name)
-            if not is_pkg_change_allowed(pkg, blacklist, whitelist):
+            if not is_pkg_change_allowed(pkg,
+                                         cache.blacklist,
+                                         cache.whitelist,
+                                         cache.strict_whitelist):
                 return ("pkg %s is blacklisted or is not whitelisted"
                         % pkg.name)
             # check if the package is unsafe to upgrade unattended
@@ -1044,7 +1173,10 @@ def get_md5sum_for_file_installed(conf_file, prefix):
     # type: (str, str) -> str
     try:
         with open(prefix + conf_file, 'rb') as fb:
-            return apt_pkg.md5sum(fb)
+            for hash_string in apt_pkg.Hashes(fb).hashes:  # type: ignore
+                if hash_string.hashtype == 'MD5Sum':
+                    return hash_string.hashvalue
+            return None
     except IsADirectoryError:
         # the package replaces a directory wih a configuration file
         #
@@ -1375,10 +1507,8 @@ def send_summary_mail(pkgs,                 # type: List[str]
     logging.debug("mail returned: %s", ret)
 
 
-def do_install(cache,             # type: apt.Cache
+def do_install(cache,             # type: UnattendedUpgradesCache
                pkgs_to_upgrade,   # type: List[apt.Package]
-               blacklist,         # type: List[str]
-               whitelist,         # type: List[str]
                options,           # type: Options
                logfile_dpkg,      # type: str
                ):
@@ -1400,7 +1530,7 @@ def do_install(cache,             # type: apt.Cache
             # try upgrade all "pkgs" in minimal steps
             pkg_install_success = upgrade_in_minimal_steps(
                 cache, [pkg.name for pkg in pkgs_to_upgrade],
-                blacklist, whitelist, logfile_dpkg,
+                logfile_dpkg,
                 options.verbose or options.debug)
         else:
             mark_pkgs_to_upgrade(cache, pkgs_to_upgrade)
@@ -1504,16 +1634,6 @@ def _setup_logging(options):
     return mem_log
 
 
-def get_blacklist():
-    # type: () -> List[str]
-    return apt_pkg.config.value_list("Unattended-Upgrade::Package-Blacklist")
-
-
-def get_whitelist():
-    # type: () -> List[str]
-    return apt_pkg.config.value_list("Unattended-Upgrade::Package-Whitelist")
-
-
 def logged_in_users():
     # type: () -> AbstractSet[str]
     """Return a list of logged in users"""
@@ -1569,11 +1689,7 @@ def write_stamp_file():
 
 def try_to_upgrade(pkg,               # type: apt.Package
                    pkgs_to_upgrade,   # type: List[apt.Package]
-                   pkgs_kept_back,    # type: KeptPkgs
                    cache,             # type: UnattendedUpgradesCache
-                   allowed_origins,   # type: List[str]
-                   blacklist,         # type: List[str]
-                   whitelist,         # type: List[str]
                    ):
     # type: (...) -> None
     try:
@@ -1588,24 +1704,17 @@ def try_to_upgrade(pkg,               # type: apt.Package
             return
         cache._cached_candidate_pkgnames.add(pkg.name)
         cache.mark_upgrade_adjusted(pkg, from_user=not pkg.is_auto_installed)
-        if check_changes_for_sanity(cache, allowed_origins, blacklist,
-                                    whitelist, pkg):
+        if check_changes_for_sanity(cache, pkg):
             # add to packages to upgrade
             pkgs_to_upgrade.append(pkg)
-            # re-eval pkgs_kept_back as the resolver may fail to
-            # directly upgrade a pkg, but that may work during
-            # a subsequent operation, see debian bug #639840
-            pkgs_to_upgrade.extend(pkgs_kept_back.pop_upgradable(cache))
         else:
             rewind_cache(cache, pkgs_to_upgrade)
-            pkgs_kept_back.add(pkg, cache)
     except (SystemError, NoAllowedOriginError) as e:
         # can't upgrade
         logging.warning(
             _("package %s upgradable but fails to "
                 "be marked for upgrade (%s)"), pkg.name, e)
         rewind_cache(cache, pkgs_to_upgrade)
-        pkgs_kept_back.add(pkg, cache)
 
 
 def candidate_version_changed(pkg,               # type: apt.Package
@@ -1616,15 +1725,11 @@ def candidate_version_changed(pkg,               # type: apt.Package
                 'Unattended-Upgrade::Allow-downgrade', False))
 
 
-def calculate_upgradable_pkgs(cache,             # type: apt.Cache
-                              options,           # type: Options
-                              allowed_origins,   # type: List[str]
-                              blacklist,         # type: List[str]
-                              whitelist,         # type: List[str]
+def calculate_upgradable_pkgs(cache,            # type: UnattendedUpgradesCache
+                              options,          # type: Options
                               ):
-    # type: (...) -> Tuple[List[apt.Package], KeptPkgs]
+    # type: (...) -> List[apt.Package]
     pkgs_to_upgrade = []  # type: List[apt.Package]
-    pkgs_kept_back = KeptPkgs(set)
 
     # now do the actual upgrade
     for pkg in cache:
@@ -1634,23 +1739,19 @@ def calculate_upgradable_pkgs(cache,             # type: apt.Cache
                 pkg.name, getattr(pkg.candidate, "origins", [])))
 
         if (pkg.is_upgradable or candidate_version_changed(pkg)
-           and is_pkgname_in_whitelist(pkg.name, whitelist)):
+           and is_pkgname_in_whitelist(pkg.name, cache.whitelist)):
             try:
-                ver_in_allowed_origin(pkg, allowed_origins)
+                ver_in_allowed_origin(pkg, cache.allowed_origins)
             except NoAllowedOriginError:
                 continue
             try_to_upgrade(pkg,
                            pkgs_to_upgrade,
-                           pkgs_kept_back,
-                           cache,
-                           allowed_origins,
-                           blacklist,
-                           whitelist)
+                           cache)
 
     if cache.get_changes():
         cache.clear()
 
-    return pkgs_to_upgrade, pkgs_kept_back
+    return pkgs_to_upgrade
 
 
 def get_dpkg_log_content(logfile_dpkg, install_start_time):
@@ -1692,8 +1793,6 @@ def get_auto_removable(cache):
 def is_autoremove_valid(cache,           # type: UnattendedUpgradesCache
                         pkgname,         # type: str
                         auto_removable,  # type: AbstractSet[str]
-                        blacklist,       # type: List[str]
-                        whitelist        # type: List[str]
                         ):
     # type: (...) -> bool
     changes = cache.get_changes()
@@ -1702,7 +1801,8 @@ def is_autoremove_valid(cache,           # type: UnattendedUpgradesCache
         return True
     pkgnames = {pkg.name for pkg in changes}
     for pkg in changes:
-        if not is_pkg_change_allowed(pkg, blacklist, whitelist):
+        if not is_pkg_change_allowed(pkg, cache.blacklist, cache.whitelist,
+                                     cache.strict_whitelist):
             logging.warning(
                 _("Keeping the following auto-removable package(s) because "
                   "they include %s which is set to be kept unmodified: %s"),
@@ -1740,12 +1840,10 @@ def is_autoremove_valid(cache,           # type: UnattendedUpgradesCache
     return True
 
 
-def do_auto_remove(cache,             # type: apt.Cache
+def do_auto_remove(cache,             # type: UnattendedUpgradesCache
                    auto_removable,    # type: AbstractSet[str]
                    logfile_dpkg,      # type: str
                    minimal_steps,     # type: bool
-                   blacklist,         # type: List[str]
-                   whitelist,         # type: List[str]
                    verbose=False,     # type: bool
                    dry_run=False      # type: bool
                    ):
@@ -1765,8 +1863,7 @@ def do_auto_remove(cache,             # type: apt.Cache
             if pkgname in pkgs_removed:
                 continue
             cache[pkgname].mark_delete()
-            if not is_autoremove_valid(cache, pkgname, auto_removable,
-                                       blacklist, whitelist):
+            if not is_autoremove_valid(cache, pkgname, auto_removable):
                 # this situation can occur when removing newly unused packages
                 # would also remove old unused packages which are not set
                 # for removal, thus getting there is not handled as an error
@@ -1785,8 +1882,7 @@ def do_auto_remove(cache,             # type: apt.Cache
     else:
         for pkgname in auto_removable:
             cache[pkgname].mark_delete()
-        if is_autoremove_valid(cache, "", auto_removable, blacklist,
-                               whitelist):
+        if is_autoremove_valid(cache, "", auto_removable):
             # do it in one step
             if not dry_run:
                 res, error = cache_commit(cache, logfile_dpkg, verbose)
@@ -1846,7 +1942,7 @@ def is_update_day():
     return False
 
 
-def update_kept_packages(kept_pkgs, kept_file):
+def update_kept_pkgs_file(kept_pkgs, kept_file):
     # type: (DefaultDict[str, List[str]], str) -> None
     if kept_pkgs:
         pkgs_all_origins = set()
@@ -1892,8 +1988,8 @@ def main(options, rootdir="/"):
 
         if res.success and res.result_str:
             # complete, successful run
-            update_kept_packages(res.pkgs_kept_back,
-                                 os.path.join(rootdir, KEPT_PACKAGES_FILE))
+            update_kept_pkgs_file(res.pkgs_kept_back,
+                                  os.path.join(rootdir, KEPT_PACKAGES_FILE))
 
         if res.result_str and not options.dry_run:
             # there is some meaningful result which is worth an email
@@ -1991,24 +2087,9 @@ def run(options,             # type: Options
         syslog.syslog(_("Not running on the development release."))
         logging.info(_("Not running on the development release."))
         return UnattendedUpgradesResult(True)
-    # format (origin, archive), e.g. ("Ubuntu","dapper-security")
-    allowed_origins = get_allowed_origins()
-
-    # pkgs that are (for some reason) not safe to install
-    blacklist = get_blacklist()
-    logging.info(_("Initial blacklist: %s"),
-                 " ".join(blacklist))
-
-    # install only these packages regardless of other upgrades available
-    whitelist = get_whitelist()
-    logging.info(_("Initial whitelist: %s"),
-                 " ".join(whitelist))
 
     logging.info(_("Starting unattended upgrades script"))
 
-    # display available origin
-    logging.info(_("Allowed origins are: %s"), ", ".join(allowed_origins))
-
     # check and get lock
     try:
         apt_pkg.pkgsystem_lock()
@@ -2042,8 +2123,7 @@ def run(options,             # type: Options
 
     # get a cache
     try:
-        cache = UnattendedUpgradesCache(rootdir=rootdir,
-                                        allowed_origins=allowed_origins)
+        cache = UnattendedUpgradesCache(rootdir=rootdir)
     except SystemError as error:
         print(_("Apt returned an error, exiting"))
         print(_("error message: %s") % error)
@@ -2074,8 +2154,7 @@ def run(options,             # type: Options
     auto_removable = get_auto_removable(cache)
 
     # find out about the packages that are upgradable (in an allowed_origin)
-    pkgs_to_upgrade, pkgs_kept_back = calculate_upgradable_pkgs(
-        cache, options, allowed_origins, blacklist, whitelist)
+    pkgs_to_upgrade = calculate_upgradable_pkgs(cache, options)
     pkgs_to_upgrade.sort(key=lambda p: p.name)
     pkgs = [pkg.name for pkg in pkgs_to_upgrade]
     logging.debug("pkgs that look like they should be upgraded: %s"
@@ -2121,6 +2200,7 @@ def run(options,             # type: Options
     if dpkg_conffile_prompt():
         # now check the downloaded debs for conffile conflicts and build
         # a blacklist
+        conffile_blacklist = []  # type: List[str]
         for item in fetcher.items:
             logging.debug("%s" % item)
             if item.status == item.STAT_ERROR:
@@ -2146,8 +2226,8 @@ def run(options,             # type: Options
                     "APT::Get::AllowUnauthenticated", False):
                 logging.debug("%s is blacklisted because it is not trusted")
                 pkg_name = pkgname_from_deb(item.destfile)
-                if not is_pkgname_in_blacklist(pkg_name, blacklist):
-                    blacklist.append("%s$" % re.escape(pkg_name))
+                if not is_pkgname_in_blacklist(pkg_name, cache.blacklist):
+                    conffile_blacklist.append("%s$" % re.escape(pkg_name))
             if not is_deb(item.destfile):
                 logging.debug("%s is not a .deb file" % item)
                 continue
@@ -2168,18 +2248,18 @@ def run(options,             # type: Options
                                   "needs to be upgraded manually"),
                                 pkgname_from_deb(item.destfile))
                 pkg_name = pkgname_from_deb(item.destfile)
-                if not is_pkgname_in_blacklist(pkg_name, blacklist):
-                    blacklist.append("%s$" % re.escape(pkg_name))
-                pkgs_kept_back.add(cache[pkg_name], cache)
+                if not is_pkgname_in_blacklist(pkg_name, cache.blacklist):
+                    conffile_blacklist.append("%s$" % re.escape(pkg_name))
                 pkg_conffile_prompt = True
 
         # redo the selection about the packages to upgrade based on the new
         # blacklist
-        logging.debug("blacklist: %s" % blacklist)
-        # whitelist
-        logging.debug("whitelist: %s" % whitelist)
+        logging.debug("Packages blacklist due to conffile prompts: %s"
+                      % conffile_blacklist)
         # find out about the packages that are upgradable (in a allowed_origin)
-        if len(blacklist) > 0 or len(whitelist) > 0:
+        if len(conffile_blacklist) > 0:
+            cache.apply_pinning(cache.pinning_from_regex_list(
+                conffile_blacklist, NEVER_PIN))  # type: ignore
             old_pkgs_to_upgrade = pkgs_to_upgrade[:]
             pkgs_to_upgrade = []
             for pkg in old_pkgs_to_upgrade:
@@ -2187,14 +2267,9 @@ def run(options,             # type: Options
                               (pkg.name))
                 cache.mark_upgrade_adjusted(
                     pkg, from_user=not pkg.is_auto_installed)
-                if check_changes_for_sanity(cache,
-                                            allowed_origins,
-                                            blacklist,
-                                            whitelist):
+                if check_changes_for_sanity(cache):
                     pkgs_to_upgrade.append(pkg)
                 else:
-                    if not (pkg.name in pkgs_kept_back):
-                        pkgs_kept_back.add(pkg, cache)
                     logging.info(_("package %s not upgraded"), pkg.name)
                     cache.clear()
                     for pkg2 in pkgs_to_upgrade:
@@ -2231,7 +2306,7 @@ def run(options,             # type: Options
              kernel_pkgs_removed,
              kernel_pkgs_kept_installed) = do_auto_remove(
                 cache, auto_removable_kernel_pkgs, logfile_dpkg,
-                options.minimal_upgrade_steps, blacklist, whitelist,
+                options.minimal_upgrade_steps,
                 options.verbose or options.debug, options.dry_run)
             auto_removable = get_auto_removable(cache)
 
@@ -2244,15 +2319,17 @@ def run(options,             # type: Options
 
     # exit if there is nothing to do and nothing to report
     if (len(pending_autoremovals) == 0
-            and len(pkgs_to_upgrade) == 0
-            and len(pkgs_kept_back) == 0):
+            and len(pkgs_to_upgrade) == 0):
         logging.info(_("No packages found that can be upgraded unattended "
                        "and no pending auto-removals"))
+
+        pkgs_kept_back = cache.find_kept_packages(options.dry_run)
         return UnattendedUpgradesResult(
             kernel_pkgs_remove_success,
             _("No packages found that can be upgraded unattended and no "
               "pending auto-removals"),
             pkgs_removed=kernel_pkgs_removed,
+            pkgs_kept_back=pkgs_kept_back,
             pkgs_kept_installed=kernel_pkgs_kept_installed,
             update_stamp=True)
 
@@ -2281,8 +2358,6 @@ def run(options,             # type: Options
         # do install
         pkg_install_success = do_install(cache,
                                          pkgs_to_upgrade,
-                                         blacklist,
-                                         whitelist,
                                          options,
                                          logfile_dpkg)
     # Was the overall run succesful: only if everything installed
@@ -2313,7 +2388,7 @@ def run(options,             # type: Options
          pkgs_removed,
          pkgs_kept_installed) = do_auto_remove(
             cache, auto_removals, logfile_dpkg, options.minimal_upgrade_steps,
-            blacklist, whitelist, options.verbose or options.debug,
+            options.verbose or options.debug,
             options.dry_run)
         successful_run = successful_run and pkg_remove_success
     # the user wants *only new* auto-removals to be removed
@@ -2326,7 +2401,7 @@ def run(options,             # type: Options
          pkgs_removed,
          pkgs_kept_installed) = do_auto_remove(
             cache, auto_removals, logfile_dpkg, options.minimal_upgrade_steps,
-            blacklist, whitelist, options.verbose or options.debug,
+            options.verbose or options.debug,
             options.dry_run)
         successful_run = successful_run and pkg_remove_success
 
@@ -2342,6 +2417,7 @@ def run(options,             # type: Options
             and pkg_install_success):
         clean_downloaded_packages(fetcher)
 
+    pkgs_kept_back = cache.find_kept_packages(options.dry_run)
     return UnattendedUpgradesResult(
         successful_run, _("All upgrades installed"), pkgs,
         pkgs_kept_back,
